<?php
/** 
*
* @package phpBB3
* @version $Id: smartfeed.php, v 2.2.27 Production 2015/01/04 mark@phpbbservices.com Exp $
* @copyright (c) Mark D. Hamill (mark@phpbbservices.com, http://phpbbservices.com) 
* @license http://opensource.org/licenses/gpl-2.0.php GNU Public License 
*
*/

/**
* @ignore
*/

// Written by Mark D. Hamill, mark@phpbbservices.com, http://phpbbservices.com
// This software is designed to work with phpBB Version 3.0.12

// This is the program that creates the RSS or Atom compliant newsfeed that will also include protected
// forums. An enhanced and slightly modified version of the the FeedCreator class, v1.8.0-dev was used 
// to generate the various feeds. The modified version can be found at http://feedcreator.org/. In addition 
// SimplePie is used to incorporate external feeds. See http://simplepie.org/.

// A companion program should be run first: smartfeed_url.php. This creates a URL that is
// used to invoke the feed. Use the URL created by this program in your newsreader. Do NOT try to run 
// this program without arguments!

// Set up the phpBB environment we need 
define('IN_PHPBB', true);
$phpbb_root_path = (defined('PHPBB_ROOT_PATH')) ? PHPBB_ROOT_PATH : './';
$phpEx = substr(strrchr(__FILE__, '.'), 1);
include($phpbb_root_path . 'common.' . $phpEx);

if (!defined('SMARTFEED_NO_LIMIT_VALUE'))	// Test to see if a Smartfeed constant exists
{
	// It doesn't so let's load them
	include ($phpbb_root_path . 'includes/constants_smartfeed.' . $phpEx);
}

// We need to start a session to take advantage of the benefits it gives, like access to common function, sessions, etc.
$user->session_begin();
$auth->acl($user->data);

// Note: one limitation of adding the smartfeed language file is that it inherits the guest default language 
// unless the user is authenticated automatically, such as via cookie authentication. It behooves the board administrator
// to make sure he has a language specific version of all Smartfeed language files for each language allowed on the board. 
$user->setup('mods/smartfeed'); 

// If the board is currently disabled, feeds should also be disabled, don't ya think?
if ($config['board_disable'])
{
	handle_error('SMARTFEED_BOARD_DISABLED');
}

// Set up some variables
$registered_user = false;
$allowed_user_types = array(USER_NORMAL, USER_FOUNDER); // Allowed users are Normal and Founder. Others (Inactive, Ignore) can only get a public feed.
$board_url = generate_board_url() . '/';
$true_false_array = array(0, 1);
$lastvisit = false;

// ------ First, report errors for any inconsistent URL errors ------ //

// Arguments are required when running smartfeed.php
$query_string = $user->page['query_string'];
if (strlen($query_string) == 0)
{
	$feed_type = $config['sf_rss_20_value'];
	handle_error('SMARTFEED_NO_ARGUMENTS');
}

// Validate the feed type. Atom 1.0, RSS 2.0 and RSS 1.0 are the only feed types allowed.
$feed_type = request_var('feed_type','NONE');
if (($feed_type == 'NONE') || !(($feed_type == $config['sf_atom_10_value']) || 
	($feed_type == $config['sf_rss_20_value']) || 
	($feed_type == $config['sf_rss_10_value'])))
{
	// Bad feed_type, trigger error
	handle_error('SMARTFEED_FEED_TYPE_ERROR');
}

// Get the user id. A public user won't be identified as a user in the URL.
$user_id = request_var('u', ANONYMOUS);

// Get the encrypted password. When decrypted it is still encoded md5 as it should also be in the database
$encrypted_pswd = request_var('e', 'NONE');

// Determine if this is a public request. If so only posts in public forums will be shown in the feed.
if ($user_id != ANONYMOUS && $encrypted_pswd != 'NONE')
{
	// Feed privileges are dependent upon the auth_method. This code makes this program consistent with smartfeed_url.php
	if (($config['auth_method'] == 'apache') && ($config['sf_apache_htaccess_enabled'] == 0))
	{
		handle_error('SMARTFEED_APACHE_AUTHENTICATION_WARNING_REG');
	}
	$registered_user = true;
}
else if (!(($user_id == ANONYMOUS) && ($encrypted_pswd == 'NONE')))
{
	// Logically if only the u or the e parameter is present, the URL is inconsisent, so generate an error.
	if ($user_id == ANONYMOUS)
	{
		handle_error('SMARTFEED_NO_U_ARGUMENT');
	}
	if ($encrypted_pswd == 'NONE')
	{
		handle_error('SMARTFEED_NO_E_ARGUMENT');
	}
}

// Get the limit parameter. It limits the size of the newsfeed to a point in time from the present, either a day/hour/minute interval, no limit
// or the time since the user's last visit. It should always exist. If it doesn't exist or the value passed is invalid, trigger an error.
// Please note that no message older than SMARTFEED_DEFAULT_FETCH_TIME_LIMIT can actually be retrieved to avoid huge 
// newsfeeds and excessive strain on the database.
$time_limit = request_var('limit', 'NONE');
if ($registered_user && (!(in_array($time_limit, $smartfeed['SMARTFEED_TIME_LIMIT_REGISTERED']))))
{
	handle_error('SMARTFEED_LIMIT_FORMAT_ERROR');
}
if (!$registered_user && (!(in_array($time_limit, $smartfeed['SMARTFEED_TIME_LIMIT_UNREGISTERED']))))
{
	handle_error('SMARTFEED_LIMIT_FORMAT_ERROR');
}

// Validate the sort_by parameter, which should always be present
$sort_by = request_var('sort_by','NONE');
if (($sort_by == 'NONE') || (!(in_array($sort_by, $smartfeed['SMARTFEED_SORT_BY_VALUES']))))
{
	// Bad sort_by, trigger error
	handle_error('SMARTFEED_SORT_BY_ERROR');
}

// Validate the firstpostonly parameter.
$first_post_only = request_var('firstpostonly',0);
if (!in_array($first_post_only, $true_false_array))
{
	handle_error('SMARTFEED_FIRST_POST_ONLY_ERROR');
}

// Check for count limit parameter. It is not required, but if present should be a positive number only. The value must
// be less than or equal to $config['sf_max_items']. But if $config['sf_max_items'] == 0 then any positive whole number is allowed.
// If not present the count limit is $config['sf_max_items'] if positive, or unlimited if this value is zero.
$count_limit = request_var('count_limit',0);
if ($count_limit < 0)
{
	handle_error('SMARTFEED_COUNT_LIMIT_ERROR');
}
if (($config['sf_max_items'] > 0) && ($count_limit > $config['sf_max_items']))
{
	handle_error('SMARTFEED_COUNT_LIMIT_MAX_ERROR');
}
if (($config['sf_max_items'] > 0) && ($count_limit == 0))
{
	$count_limit = $config['sf_max_items'];
}

// Validate the maximum size requested for the number of words in a post
$max_word_size = request_var('max_word_size',0);
if ($max_word_size < 0)
{
	// Bad max_word_size, trigger error
	handle_error('SMARTFEED_MAX_WORD_SIZE_ERROR');
}
	
// Validate the minimum size requested for the number of words in a post
$min_word_size = request_var('min_word_size',0);
if ($min_word_size < 0)
{
	handle_error('SMARTFEED_MIN_WORD_SIZE_ERROR');
}
	
// Validate the feed_style parameter.
$feed_style = request_var('feed_style','NONE');
if (($sort_by == 'NONE') || (!(in_array($feed_style, $smartfeed['SMARTFEED_FEED_STYLE_VALUES']))))
{
	handle_error('SMARTFEED_STYLE_ERROR');
}

if ($registered_user)
{
	// Validate the removemine parameter, if present
	$remove_mine = request_var('removemine',0);
	if (!in_array($remove_mine, $true_false_array))
	{
		// Bad sort_bye, trigger error
		handle_error('SMARTFEED_REMOVE_MINE_ERROR');
	}

	// Validate the private messages switch
	$show_pms = request_var('pms',0);
	if (!in_array($show_pms, $true_false_array))
	{
		// Bad parameter value for private messages
		handle_error('SMARTFEED_BAD_PMS_VALUE');
	}

	// Validate the mark read private messages switch
	$mark_read_pms = request_var('mark',0);
	if (!in_array($mark_read_pms, $true_false_array))
	{
		// Bad parameter value for private messages
		handle_error('SMARTFEED_BAD_MARK_READ_VALUE');
	}

	// Validate the bookmarks switch
	$bookmarks = request_var('bookmarks',0);
	if (!in_array($bookmarks, $true_false_array))
	{
		// Bad parameter value for private messages
		handle_error('SMARTFEED_BAD_BOOKMARKS_VALUE');
	}

	// Validate the filter_foes parameter.
	$filter_foes = request_var('filter_foes',0);
	if (!in_array($filter_foes, $true_false_array))
	{
		handle_error('SMARTFEED_FILTER_FOES_ERROR');
	}
	
	// Validate the lastvisit parameter.
	$lastvisit = request_var('lastvisit',0);
	if (!in_array($lastvisit, $true_false_array))
	{
		handle_error('SMARTFEED_LASTVISIT_ERROR');
	}
}

// ------ Second, authenticate the user ------ //

$sql = 'SELECT user_id, user_password, user_smartfeed_key, user_topic_sortby_type, user_topic_sortby_dir, 
			user_post_sortby_type, user_post_sortby_dir, user_lastvisit, user_type
		FROM ' . USERS_TABLE . " WHERE user_id = $user_id";
if ($user_id != ANONYMOUS)
{
	$sql .= ' AND ' . $db->sql_in_set('user_type', $allowed_user_types); // Robots and inactive members are not allowed to get into restricted forums
}

$result = $db->sql_query($sql);
$rowset = $db->sql_fetchrowset($result);

if (sizeof($rowset) == 0)
{
	handle_error('SMARTFEED_USER_ID_DOES_NOT_EXIST');
}

// Make sure user_id exists in database and has normal or founder status
$row = reset($rowset);

// Save the user variables, although only the first is unneeded for guests.
$user_smartfeed_key = $row['user_smartfeed_key'];
$user_topic_sortby_type = $row['user_topic_sortby_type']; 
$user_topic_sortby_dir = $row['user_topic_sortby_dir']; 
$user_post_sortby_type = $row['user_post_sortby_type']; 
$user_post_sortby_dir = $row['user_post_sortby_dir'];

// These other variables are only used by registered users
$user_password = $row['user_password'];
$user_lastvisit = $row['user_lastvisit'];

$db->sql_freeresult($result); // Query be gone!

// Get any external newsfeeds
$ext_feeds = explode("\n",trim($config['sf_external_feeds']));

if ($registered_user)
{
	
	// Decrypt password using the user_smartfeed_key column in the phpbb_users table. This should have gotten created 
	// the first time smartfeed_url.php was run by this user. Note the encoded password is typically md5. There should not be
	// a clear text password in the database.
	if (strlen($user_smartfeed_key) > 0)
	{
		$encoded_pswd = decrypt($encrypted_pswd, $user_smartfeed_key);
	}
	else
	{
		// If the $user_smartfeed_key is null, password cannot be decrypted. This avoids PHP Notice statements in TripleDES.php
		handle_error('SMARTFEED_BAD_PASSWORD_ERROR');
	}
				
	// If IP Authentication was enabled, the encoded password is to the left of the ~ and the IP to the right of the ~
	$tilde = strpos($encoded_pswd, '~');
	if (($tilde == 0) && ($config['sf_require_ip_authentication'] == '1'))
	{
		handle_error('SMARTFEED_IP_AUTH_ERROR');
	}
	
	if ($tilde > 0)
	{
		// Since a tilde is present, authenticate the client IP by comparing it with the IP embedded in the "e" parameter
		$auth_ip = substr($encoded_pswd,$tilde+1);
		$encoded_pswd = substr($encoded_pswd,0,$tilde);
		$client_ip_parts = explode('.', $user->ip);
		$source_ip_parts = explode('.', $auth_ip);
		// Show error message if requested from incorrect range of IP addresses
		if (!(
				($client_ip_parts[0] == $source_ip_parts[0]) && 
				($client_ip_parts[1] == $source_ip_parts[1]) &&
				(($client_ip_parts[2] == $source_ip_parts[2]) || ($source_ip_parts[2] == '*'))
			))
		{
			handle_error('SMARTFEED_IP_AUTH_ERROR');
		}
	}

	if ($encoded_pswd != $user_password)
	{
		handle_error('SMARTFEED_BAD_PASSWORD_ERROR');
	} 

}

// If requested to fetch new posts since the user's last visit, we need to get the user_lastvisit date from the database. Of course this makes
// no sense if the user is Anonymous. Cookie logic in the phpBB 2 version has been removed to simplify things. Most newsreaders won't store cookies
// anyway.

switch($config['sf_default_fetch_time_limit'])
{
	case 0: // Admin has set no limit to how far back posts can be fetched
		switch ($time_limit)
		{
			case SMARTFEED_SINCE_LAST_VISIT:
				$date_limit = $user_lastvisit;
				$date_limit_sql = ' AND p.post_time > ' . $date_limit;
			break;
			case SMARTFEED_NO_LIMIT_VALUE:
				$date_limit = 0;
				$date_limit_sql = '';
			break;
			default:
				$date_limit = $smartfeed['SMARTFEED_TIME_LIMIT_SECONDS'][$time_limit];
				$date_limit_sql = ' AND p.post_time > ' . $date_limit;
			break;
		}
	break;
		
	default: // Admin has set a limit to how far back posts can be fetched
		switch ($time_limit)
		{
			case SMARTFEED_SINCE_LAST_VISIT:
				$date_limit = max($user_lastvisit, time() - ($config['sf_default_fetch_time_limit'] * 3600));
				$date_limit_sql = ' AND p.post_time > ' . $date_limit;
			break;
			case SMARTFEED_NO_LIMIT_VALUE:
				$date_limit = time() - ($config['sf_default_fetch_time_limit'] * 3600);
				$date_limit_sql = ' AND p.post_time > ' . $date_limit;
			break;
			default:
				$date_limit = max($smartfeed['SMARTFEED_TIME_LIMIT_SECONDS'][$time_limit], time() - ($config['sf_default_fetch_time_limit'] * 3600));
				$date_limit_sql = ' AND p.post_time > ' . $date_limit;
			break;
		}
	break;
}

// ------ Third, create a list of forums that will be fetched ------ //

$fetched_forum_str = '';

if ($registered_user && $bookmarks)
{

	// When selecting bookmarked topics only, we can safely ignore the logic constraining the user to read only 
	// from certain forums. Instead we will create the SQL to get the bookmarked topics, if any, hijacking the 
	// $fetched_forum_str variable since it is convenient
	
	$bookmarked_topics = array();
	$sql = 'SELECT t.topic_id
			FROM ' . USERS_TABLE . ' u, ' . BOOKMARKS_TABLE . ' b, ' . TOPICS_TABLE . " t
			WHERE u.user_id = b.user_id AND b.topic_id = t.topic_id 
				AND t.topic_last_post_time > $date_limit
				AND b.user_id = $user_id";
	$result = $db->sql_query($sql);
	while ($row = $db->sql_fetchrow($result))
	{
		$bookmarked_topics[] = intval($row['topic_id']);
	}
	$db->sql_freeresult($result);
	if (sizeof($bookmarked_topics) > 0)
	{
		$fetched_forums_str = ' AND ' . $db->sql_in_set('t.topic_id', $bookmarked_topics);
	}
	else
	{
		// Logically, if there are no bookmarked topics for this $user_id then there will be nothing in the feed.
		// Send a message to this effect in the feed.
		handle_error('SMARTFEED_NO_BOOKMARKS');
	}

}
else
{

	// Getting a list of allowed forums is now much simpler now that I know about the acl_raw_data_single_user function. 
	
	// We need to know which auth_option_id corresponds to the forum read privilege (f_read) and forum list (f_list) privilege.
	$auth_options = array('f_read', 'f_list');
	$sql = 'SELECT auth_option, auth_option_id
			FROM ' . ACL_OPTIONS_TABLE . '
			WHERE ' . $db->sql_in_set('auth_option', $auth_options);
	$result = $db->sql_query($sql);
	while ($row = $db->sql_fetchrow($result))
	{
		if ($row['auth_option'] == 'f_read')
		{
			$read_id = $row['auth_option_id'];
		}
		if ($row['auth_option'] == 'f_list')
		{
			$list_id = $row['auth_option_id'];
		}
	}
	$db->sql_freeresult($result); // Query be gone!

	// Now let's get this user's forum permissions. Note that non-registered, robots etc. get a list of public forums
	// with read permissions.
	
	$allowed_forums = array();
	
	$forum_array = $auth->acl_raw_data_single_user($user_id);
	foreach ($forum_array as $key => $value)
	{
	
		foreach ($value as $auth_option_id => $auth_setting)
		{
			if ($auth_option_id == $read_id)
			{
				if (($auth_setting == 1) && check_all_parents($key))
				{
					$allowed_forums[] = $key;
				}
			}
		}
		
	}
	
	if (sizeof($allowed_forums) == 0)
	{
		// If this user cannot retrieve ANY forums, this suggests that this board is tightly locked down to members only,
		// or every member must belong to a user group or have special forum permissions
		handle_error('SMARTFEED_NO_ACCESSIBLE_FORUMS');
	}
	
	// Get the requested forums. If none are listed, user wants all forums for which they have read access.
	$requested_forums = array();
	$params = explode('&', $query_string);
	$required_forums_only = false;
	foreach ($params as $item)
	{
		if ($item == 'forum=-1')
		{
			// This is an unusual case and it means that no forums were selected but there are required forums.
			// In this case the feed is restricted to returning content ONLY for required forums.
			$required_forums_only = true;
			break;
		}
		if (substr($item,0,5) == 'forum')
		{
			$requested_forums[] = substr($item,6);
		}
	}
	
	// To capture global announcements when forums are specified, we have to add the pseudo-forum with a forum_id = 0.
	if (sizeof($requested_forums) > 0)
	{
		$requested_forums[] = '0';
	}
	
	// Sort requested forums by forum_id and ensure there are no duplicates
	asort($requested_forums);
	$requested_forums = array_unique($requested_forums);
	
	// The forums that will be fetched is the array intersection of the requested and allowed forums. This prevents hacking
	// the URL to get feeds a user is not supposed to get. If no forums are specified on the URL field
	// then all forums that this user is authorized to access is assumed.
	
	if (!$required_forums_only)
	{
		$fetched_forums = (sizeof($requested_forums) > 0) ? array_intersect($allowed_forums, $requested_forums): $allowed_forums;
	}
	else
	{
		$fetched_forums = $requested_forums;
	}
	
	// Add in any required forums
	if (strlen($config['sf_include_forums']) > 0)
	{
		$fetched_forums = array_merge($fetched_forums, $required_forums);
	}
	
	// Remove any prohibited forums
	$excluded_forums = (isset($config['sf_exclude_forums'])) ? explode(',',$config['sf_exclude_forums']) : array();
	if (sizeof($excluded_forums) > 0)
	{
		$fetched_forums = array_diff($fetched_forums, $excluded_forums);
	}
	$fetched_forums = array_unique($fetched_forums);

	// Create a SQL fragment to return posts from the correct forums
	if (sizeof($fetched_forums) > 0)
	{
		$fetched_forums_str = ' AND ' . $db->sql_in_set('p.forum_id', $fetched_forums);
	}
	else
	{
		// If there are no forums to fetch, this will result in an empty newsfeed. 
		handle_error('SMARTFEED_NO_FORUMS_ACCESSIBLE');
	}

}

// ------ Fourth, create the snippets of SQL used to return the set of posts ------ //

// Create the SQL stub for the sort order
switch($sort_by)
{
	case SMARTFEED_BOARD:
		$topic_asc_desc = ($user_topic_sortby_dir == 'd') ? 'DESC' : '';
		switch($user_topic_sortby_type)
		{
			case 'a':
				$order_by_sql = "t.topic_first_poster_name $topic_asc_desc, ";
			break;
			case 't':
				$order_by_sql = "t.topic_last_post_time $topic_asc_desc, ";
			break;
			case 'r':
				$order_by_sql = "t.topic_replies $topic_asc_desc, ";
			break;
			case 's':
				$order_by_sql = "t.topic_title $topic_asc_desc, " ; 
			break;
			case 'v':
				$order_by_sql = "t.topic_views $topic_asc_desc, ";
			break;
		}
		$post_asc_desc = ($user_post_sortby_dir == 'd') ? 'DESC' : '';
		switch($user_post_sortby_type)
		{
			case 'a':
				$order_by_sql .= "u.username_clean $post_asc_desc";
			break;
			case 't':
				$order_by_sql .= "p.post_time $post_asc_desc";
			break;
			case 's':
				$order_by_sql .= "p.post_subject $post_asc_desc" ; 
			break;
		}
	break;
	case SMARTFEED_STANDARD:
		$order_by_sql = 'f.left_id, f.right_id, t.topic_last_post_time, p.post_time';
	break;
	case SMARTFEED_STANDARD_DESC:
		$order_by_sql = 'f.left_id, f.right_id, t.topic_last_post_time, p.post_time DESC';
	break;
	case SMARTFEED_POSTDATE:
		$order_by_sql = 'p.post_time';
	break;
	case SMARTFEED_POSTDATE_DESC:
		$order_by_sql = 'p.post_time DESC';
	break;
}

$new_topics_sql = '';
$topics_posts_join_sql = 'p.topic_id = t.topic_id';

// Create the first_post_only SQL stubs
if ($first_post_only)
{
	$new_topics_sql = " AND t.topic_time > $date_limit ";
	$topics_posts_join_sql = ' t.topic_first_post_id = p.post_id AND t.forum_id = f.forum_id';
}

// Create SQL to remove your posts from the feed
$remove_mine_sql = '';
if ($registered_user && ($remove_mine == 1))
{
	$remove_mine_sql = " AND p.poster_id <> $user_id ";
}

// Create SQL to remove your foes from the feed
$filter_foes_sql = '';
$foes = array();
if ($registered_user && ($filter_foes == 1))
{

	// Fetch your foes
	$sql = 'SELECT zebra_id 
			FROM ' . ZEBRA_TABLE . "
			WHERE user_id = $user_id AND foe = 1";
	$result = $db->sql_query($sql);
	while ($row = $db->sql_fetchrow($result))
	{
		$foes[] = (int) $row['zebra_id'];
	}
	$db->sql_freeresult($result);

	if (sizeof($foes) > 0)
	{
		$filter_foes_sql = ' AND ' . $db->sql_in_set('p.poster_id', $foes, true);
	}
	
}

// At last, construct the SQL to return the relevant posts
			
$sql_array = array(
	'SELECT'	=> 'f.*, t.*, p.*, u.*, tt.mark_time AS topic_mark_time, ft.mark_time AS forum_mark_time',

	'FROM'		=> array(
		POSTS_TABLE => 'p',
		USERS_TABLE => 'u',
		TOPICS_TABLE => 't'),

	'WHERE'		=> "$topics_posts_join_sql
				AND p.poster_id = u.user_id
				$date_limit_sql
				$fetched_forums_str
				$new_topics_sql
				$remove_mine_sql
				$filter_foes_sql
				AND p.post_approved = 1",

	'ORDER_BY'	=> $order_by_sql
);

$sql_array['LEFT_JOIN'] = array(
	array(
		'FROM'	=> array(FORUMS_TABLE => 'f', TOPICS_TABLE => 't'),
		'ON'	=> 't.forum_id = f.forum_id'
	),
	array(
		'FROM'   => array(TOPICS_TRACK_TABLE => 'tt', TOPICS_TABLE => 't'),
		'ON'   => "t.topic_id = tt.topic_id AND tt.user_id = $user_id"
	),
	array(
		'FROM'   => array(FORUMS_TRACK_TABLE => 'ft', FORUMS_TABLE => 'f'),
		'ON'   => "f.forum_id = ft.forum_id AND ft.user_id = $user_id"
	)
);

$sql = $db->sql_build_query('SELECT', $sql_array);

// Now finally, let's fetch the actual posts to be placed in this newsfeed
$result = $db->sql_query_limit($sql, $count_limit); // Execute the SQL to retrieve the relevant posts. Note, if $count_limit is 0 then there is no limit on the rows returned
$rowset = $db->sql_fetchrowset($result); // Get all the posts as a set

if ($registered_user && $show_pms)
{

	// If there are any unread private messages, they are fetched separately and passed as a rowset to publish_feed.
	$pm_sql = 	'SELECT *
				FROM ' . PRIVMSGS_TO_TABLE . ' pt, ' . PRIVMSGS_TABLE . ' pm, ' . USERS_TABLE . " u
				WHERE pt.msg_id = pm.msg_id
					AND pt.author_id = u.user_id
					AND pt.user_id = $user_id
					AND (pm_unread = 1 OR pm_new = 1)";
	$pm_result = $db->sql_query($pm_sql);
	$pm_rowset = $db->sql_fetchrowset($pm_result);

}
else
{
	$pm_result = NULL;
	$pm_rowset = NULL;
}

publish_feed($rowset, $pm_rowset, $ext_feeds); // Publish the set of rows (posts and private messages) as feed items
$db->sql_freeresult($result);
$db->sql_freeresult($pm_result);

// Reset the user's last visit date on the forum, if so requested
if (isset($lastvisit))
{
	if ($lastvisit)
	{
		$sql = 'UPDATE ' . USERS_TABLE . '
					SET user_lastvisit = ' . time() . " 
					WHERE user_id = $user_id";
			
		$db->sql_query($sql);
	}
}

// Session can be killed at this point and is normally preferred so as not to skew online and guest statistics, but not if:
//	- $config['sf_show_sessions'] is true
//	- User is administrator, because they are likely testing and don't want the hassle of logging back in
//	- User just ran smartfeed_url.php, because they are likely testing too

$kill_session = true;
if ($config['sf_show_sessions'])
{
	$kill_session = false; 
}
elseif ($user->data['user_type'] == USER_FOUNDER)
{
	$kill_session = false; 
}
elseif (strpos($user->referer,'smartfeed_url') === true)
{
	$kill_session = false; 
}

// Kill the session, but only if smartfeed_url.php was not calling this program. Otherwise, the user's phpBB session will end
if ($kill_session)
{
	kill_session();
}

garbage_collection();	// This should kill any SQL connection rather than letting it hang

// All done!
exit_handler();

function kill_session ($new_session=true)
{

   // Kill the sessions we consumed. We don't want to use session_kill() because it updates user_lastvisit,
   // which we don't necessarily want to do since the user may want to visit the board later and see
   // messages in the feed as unread.
   
   global $db, $user, $config, $phpEx, $phpbb_root_path;
   
   $sql = 'DELETE FROM ' . SESSIONS_TABLE . "
		 WHERE session_id = '" . $db->sql_escape($user->session_id) . "'
			AND session_user_id = " . (int) $user->data['session_id'];
   $db->sql_query($sql);
   
   // Allow connecting logout with external auth method logout
   $method = basename(trim($config['auth_method']));
   require_once($phpbb_root_path . 'includes/auth/auth_' . $method . '.' . $phpEx);

   $method = 'logout_' . $method;
   if (function_exists($method))
   {
	  $method($this->data, $new_session);
   }

   return;
}

function handle_error($error_code)
{

	// Publish the error as a newsfeed with a single item in it.
	
	global $user, $encrypted_pswd, $user_id, $board_url, $phpEx, $feed_type, $config;
	if (!(isset($feed_type)))
	{
		$feed_type = $config['sf_rss_20_value'];
	}
	
	switch ($error_code)
	{
		case 'SMARTFEED_BAD_PASSWORD_ERROR':
			$error_msg = sprintf($user->lang['SMARTFEED_ERROR'][$error_code], $encrypted_pswd, $user_id, $board_url . "smartfeed_url.$phpEx");
		break;
		
		case 'SMARTFEED_COUNT_LIMIT_MAX_ERROR':
			$error_msg = sprintf($user->lang['SMARTFEED_ERROR'][$error_code], $config['sf_max_items']);
		break;
		
		case 'SMARTFEED_IP_AUTH_ERROR':
		case 'SMARTFEED_NO_ARGUMENTS':
		case 'SMARTFEED_NO_E_ARGUMENT':
		case 'SMARTFEED_NO_U_ARGUMENT':
		case 'SMARTFEED_USER_ID_DOES_NOT_EXIST':
			$error_msg = sprintf($user->lang['SMARTFEED_ERROR'][$error_code], $phpEx);
		break;

		default:
			$error_msg = $user->lang['SMARTFEED_ERROR'][$error_code];
		break;
	}
	
	publish_feed($error_msg);
	
	// Kill the session, but only if smartfeed_url.php was not calling this program. Otherwise, the user's phpBB session will end
	if (strpos($user->referer,'smartfeed_url') === false)
	{
		kill_session();
	}
	
	exit_handler();

}

function publish_feed($rowset, $pm_rowset = NULL, $feeds = NULL)
{

	// This function does the grunt work of changing rows in a SQL query into a newsfeed. It also handles
	// any potential errors that may have occurred by placing the error as the sole item of the feed itself. 
	
	global $board_url, $user, $feed_type, $phpEx, $config, $phpbb_root_path, $db, $registered_user, $max_word_size, $min_word_size, $feed_style, $user_id, $query_string, $mark_read_pms, $date_limit;

	include($phpbb_root_path . 'includes/SimplePie/autoloader.php'); // SimplePie is used to fetch external newsfeeds, if any
	
	if (!defined('FEEDCREATOR_VERSION'))
	{
		include ($phpbb_root_path . "includes/feedcreator.class.$phpEx");
	}

	// $allowable_tags is used with strip_tags. This removed "dangerous" tags that in the W3C's opinions should not be embedded in a feed,
	// including within a CDATA section. Any tag that is not one of these is stripped providing the feed style is Safe HTML.
	$allowable_tags = '<abbr><accept><accept-charset><accesskey><action><align><alt><axis><border><cellpadding><cellspacing><char><charoff><charset><checked><cite><class><clear><cols><colspan><color><compact><coords><datetime><dir><disabled><enctype><for><frame><headers><height><href><hreflang><hspace><id><ismap><label><lang><longdesc><maxlength><media><method><multiple><name><nohref><noshade><nowrap><prompt><readonly><rel><rev><rows><rowspan><rules><scope><selected><shape><size><span><src><start><summary><tabindex><target><title><type><usemap><valign><value><vspace><width>'; 
   	
	if (is_string($rowset)) // If the first argument is a string rather than an array, some error occurred so we need to publish the error as a newsfeed item
	{
		$error_mode = true;
		$error_msg = $rowset;	// The argument contains the error message
	}
	else 
	{
		// This is the typical case. We are publishing a newsfeed
		$error_mode = false;
	}
		
	// Initialize the feed
	$rss = new UniversalFeedCreator();
	$rss->useCached($feed_type, '', $config['sf_ttl'] * 60);
	$rss->title = html_entity_decode($config['sitename']);
	$rss->description = $config['site_desc'];
	$rss->link = ($feed_type == $config['sf_atom_10_value']) ? $board_url . "smartfeed.$phpEx?" . $query_string : $board_url . "smartfeed.$phpEx?" . htmlspecialchars($query_string);
	$rss->syndicationURL = $board_url . "smartfeed.$phpEx";
	
	// Some RSS 2.0 tags
	$rss->language = $config['sf_rfc1766_lang'];
	$rss->pubDate = time();
	$rss->ttl = $config['sf_ttl'];
	$rss->copyright = $user->lang['SMARTFEED_COPYRIGHT'];
	$rss->editor = $user->lang['SMARTFEED_EDITOR'];
	$rss->editorEmail = $user->lang['SMARTFEED_EDITOR_EMAIL'];
	$rss->webmaster = $config['sf_webmaster'];
	
	// Put an image in the feed. Typically this will be the phpBB default image
	$image = new FeedImage();
	$image->title = $rss->title; // Enhance to use default logo if not specified
	$image->url = $board_url . 'styles/' . get_default_style() . '/' . $config['sf_feed_image_path']; 
	$image->link = $rss->link;
	$image->description = strip_tags($config['site_desc']);
	$rss->image = $image;
	
	if ($error_mode)
	{
		// Place the error as a feed item
		$item = new FeedItem();
		$item->title = html_entity_decode($user->lang['SMARTFEED_ERROR_TITLE']);
		$item->description = sprintf($user->lang['SMARTFEED_ERROR_INTRODUCTION'], $board_url . "smartfeed_url.$phpEx") . $error_msg;
		$item->author = $config['board_contact'];
		$item->authorEmail = $config['board_email'];
		$item->link = $board_url . "smartfeed_url.$phpEx";
		$item->guid = $item->link;
		$rss->addItem($item);
	}
	else
	{
	
		// If there are any unread private messages, publish them first
		if (is_array($pm_rowset) && (sizeof($pm_rowset) > 0))
		{
		
			foreach ($pm_rowset as $row)
			{

				// This private message goes in the newsfeed
				if ($config['sf_new_post_notifications_only'])
				{
					$username = $user->lang['ADMINISTRATOR'];
				}
				else
				{
					$username = $row['username']; // Don't need to worry about Anonymous users for private messages, they cannot send them
				}
					
				$item = new FeedItem();
	
				// Create the title for the private message
				if ($config['sf_new_post_notifications_only'])
				{
					$title = $user->lang['SMARTFEED_NEW_PMS_NOTIFICATIONS_SHORT'];
				}
				else
				{
					$title = $user->lang['PRIVATE_MESSAGE'] . $user->lang['SMARTFEED_DELIMITER'] . $row['message_subject'] . $user->lang['SMARTFEED_DELIMITER'] . $user->lang['FROM'] . ' ' . $username;
				}
				
				$item->title = html_entity_decode(censor_text($title));
				
				// Create a number of item attributes, not all of which are necessarily in the feed (it depends on the feed type)
				$item->date = $row['message_time'];
				$item->pubDate = $item->date;
				
				if ($config['sf_new_post_notifications_only'])
				{
					$item->link = $board_url . 'ucp.' . $phpEx . '?i=pm&folder=inbox';
				}
				else
				{
					$item->link = $board_url . 'ucp.' . $phpEx . '?i=pm&mode=view&f=0&p=' . $row['msg_id'];
				}
				$item->source = $board_url;
				$item->category = $user->lang['PRIVATE_MESSAGE'];
				$item->guid = $item->link;
				$item->descriptionHtmlSyndicated = true;
				// The comment URL is not generated until the message is read, so we have to use the same URL as the view message link
				$item->comments = $board_url . 'ucp.' . $phpEx . '?i=pm&mode=view&f=0&p=' . $row['msg_id'];
				$item->author = $username;
	
				if ($feed_type==$config['sf_rss_20_value'] && !$config['sf_new_post_notifications_only']) 
				{
					// RSS 2.0 requires an email field to validate. Use a fake email address (no_email@example.com), set in feed_creator.class.php 
					// if the author field is blank) unless in user profile it says it is okay to show email address.
					$item->authorEmail = (($row['user_allow_viewemail'] == '1') && (!$config['sf_privacy_mode'])) ? $row['user_email'] : '';
				}               
				else
				{
					$item->authorEmail = '';
				}
				
				// To "dress up" the message text with bbCode, images, smilies etc., we need to use generate_text_for_display() function. 
				
				if ($config['sf_new_post_notifications_only'])
				{
					$message = $user->lang['SMARTFEED_NEW_PMS_NOTIFICATIONS_ONLY'];
				}
				else
				{
				
					$message = censor_text($row['message_text']);
					
					$user_sig = ( $row['enable_sig'] && ($row['user_sig'] != '') && $config['allow_sig'] && ($config['sf_privacy_mode'] == '0') ) ? $row['user_sig'] : '';
					
					if (($feed_style == SMARTFEED_HTML) || ($feed_style == SMARTFEED_HTMLSAFE))
					{
						$flags = (($row['enable_bbcode']) ? OPTION_FLAG_BBCODE : 0) +
							(($row['enable_smilies']) ? OPTION_FLAG_SMILIES : 0) + 
							(($row['enable_magic_url']) ? OPTION_FLAG_LINKS : 0);
							
						$message = generate_text_for_display($message, $row['bbcode_uid'], $row['bbcode_bitfield'], $flags);
			
						// Handle logic to display attachments in private messages
						if ($row['message_attachment'] > 0)
						{
							$message .= sprintf("<div class=\"box\">\n<p>%s</p>\n", $user->lang['ATTACHMENTS']);
							
							// Get all attachments
							$sql3 = 'SELECT *
								FROM ' . ATTACHMENTS_TABLE . '
								WHERE post_msg_id = ' . $row['msg_id'] . ' AND in_message = 1 
								ORDER BY attach_id';
							$result3 = $db->sql_query($sql3);
							while ($row3 = $db->sql_fetchrow($result3))
							{
								$file_size = round(($row3['filesize']/1024),2);
								// Show images, link to other attachments
								if (substr($row3['mimetype'],0,6) == 'image/')
								{
									$anchor_begin = '';
									$anchor_end = '';
									$pm_image_text = '';
									$thumbnail_parameter = '';
									$is_thumbnail = ($row3['thumbnail'] == 1) ? true : false;
									// Logic to resize the image, if needed
									if ($is_thumbnail)
									{
										$anchor_begin = sprintf("<a href=\"%s\">", $board_url . "download/file.$phpEx?id=" . $row3['attach_id']);
										$anchor_end = '</a>';
										$pm_image_text = $user->lang['SMARTFEED_POST_IMAGE_TEXT'];
										$thumbnail_parameter = '&t=1';
									}
									$message .= sprintf("%s<br /><em>%s</em> (%s %s)<br />%s<img src=\"%s\" alt=\"%s\" title=\"%s\" />%s\n<br />%s", $row3['attach_comment'], $row3['real_filename'], $file_size, $user->lang['KIB'], $anchor_begin, $board_url . "download/file.$phpEx?id=" . $row3['attach_id'] . $thumbnail_parameter, $row3['attach_comment'], $row3['attach_comment'], $anchor_end, $pm_image_text);
								}
								else
								{
									$message .= ($row3['attach_comment'] == '') ? '' : '<em>' . $row3['attach_comment'] . '</em><br />';
									$message .= 
										sprintf("<img src=\"%s\" title=\"\" alt=\"\" /> ", 
											$board_url . 'styles/' . get_default_style() . '/imageset/icon_topic_attach.gif') .
										sprintf("<b><a href=\"%s\">%s</a></b> (%s KiB)<br />",
											$board_url . "download/file.$phpEx?id=" . $row3['attach_id'], 
											$row3['real_filename'], 
											$file_size);
								}
							}
							$db->sql_freeresult($result3);
							
							$message .= '</div>';
										
						}

						$message = str_replace('<img src="./', '<img src="' . $board_url, $message);
						
						if ($user_sig != '')
						{
							$user_sig = generate_text_for_display(censor_text($user_sig), $row['user_sig_bbcode_uid'], $row['user_sig_bbcode_bitfield'], $flags);
						}
			
						$message = ($user_sig != '') ? $message . $user->lang['SMARTFEED_POST_SIGNATURE_DELIMITER'] . $user_sig : $message;
						if ($feed_style == SMARTFEED_HTMLSAFE)
						{
							$message = strip_tags($message, $allowable_tags);
						}
					}
					else
					{
						// Either Compact or Basic Style wanted
						if ($feed_style == SMARTFEED_BASIC)
						{
							$message = ($user_sig != '') ? $message . "\n\n" . $user_sig : $message;
						}
						strip_bbcode($message); 			// Remove the BBCode
						$message = strip_tags($message);	// Gets rid of any embedded HTML
						// Either condense all text or make line feeds explicit
						$message = ($feed_style == SMARTFEED_BASIC) ? nl2br($message) : str_replace("\n", ' ', $message);
					}
					
					// Handle the maximum number of words requested per PM logic
					if ($max_word_size != 0)
					{
						$message = truncate_words($message, intval($max_word_size));
					}

				}
				
				// Add the item (post) to the feed
				$item->description = $message;
				$rss->addItem($item);
				
				if ($mark_read_pms)
				{
					// Mark the private message as read
					$sql = 'UPDATE ' . PRIVMSGS_TO_TABLE . "
							SET pm_unread = 0, pm_new = 0  
							WHERE msg_id = " . $row['msg_id'] . " 
								AND user_id = $user_id
								AND author_id = " . $row['author_id'] . " 
								AND folder_id = " . $row['folder_id'];
					$db->sql_query($sql);
					// Decrement the user_unread_privmsg and user_new_privmsg count
					$sql = 'UPDATE ' . USERS_TABLE . " 
							SET user_unread_privmsg = user_unread_privmsg - 1,
								user_new_privmsg = user_new_privmsg - 1
							WHERE user_id = $user_id";
					$db->sql_query($sql);
				}
				
				if ($config['sf_new_post_notifications_only'])
				{
					// In this case we only want to show a message that there are private message, so one time thru this loop is enough.
					break;
				}
				
			}
			
		}

		// If requested to put external items at the top of the feed, do it here.
		if ($config['sf_external_feeds_top'] == 1)
		{
			publish_external_feeds($feeds, $rss);
		}
		
		$topics_array = array();
		
		// loop through the rowset, each row is an item in the feed
		foreach ($rowset as $row)
		{
			
			if (!(in_array($row['topic_id'], $topics_array)))
			{
				array_push($topics_array, $row['topic_id']);
				$new_topic = true;
			}
			else
			{
				$new_topic = false;
			}
			
			// Is this topic or forum associatd with the post being tracked by this user? Exclude the post if the topic track time or forum track time
			// is before the current date and time.
			if (((!is_null($row['forum_mark_time']) && ($date_limit !== 0) && ($row['forum_mark_time']) < $date_limit)) ||
				((!is_null($row['topic_mark_time']) && ($date_limit !== 0) && ($row['topic_mark_time']) < $date_limit)))
			{
				$include_post = false;
			}
			else
			{
				$include_post = true;
			}
			
			// Allow a post in the feed if no minimum number of words is specified by the user OR if a minimum number 
			// of words is specified by the user and it exceeds their minimum allowed number of words. Also allow if 
			// new post notifications is flagged and the topic has not already been visited. Allow it if it meets the 
			// condition above for being after a forum or topic's marked time for the user.
			
			if ((($min_word_size == 0 && !$config['sf_new_post_notifications_only']) ||
				($min_word_size != 0 && !$config['sf_new_post_notifications_only'] && truncate_words($row['post_text'], intval($max_word_size), true) >= $min_word_size)) ||
				($config['sf_new_post_notifications_only'] && $new_topic && $include_post))
			{
				// This post goes in the newsfeed

				if ($config['sf_new_post_notifications_only'])
				{
					$username = $user->lang['ADMINISTRATOR'];
				}
				else
				{
					$username = ($row['user_id'] == ANONYMOUS) ? $row['post_username'] : $row['username'];
				}
	
				$item = new FeedItem();
	
				// Create the title for the item (post)
				if ($config['sf_new_post_notifications_only'])
				{
					if ($config['sf_suppress_forum_names'])
					{
						$title = $row['topic_title'];
					}
					else
					{
						$forum_name = ($row['forum_name'] == NULL) ? $user->lang['SMARTFEED_GLOBAL_ANNOUNCEMENT'] : $row['forum_name'];
						$title = $forum_name . $user->lang['SMARTFEED_DELIMITER'] . $row['topic_title'];
					}
				}
				else
				{
					$forum_name = ($row['forum_name'] == NULL) ? $user->lang['SMARTFEED_GLOBAL_ANNOUNCEMENT'] : $row['forum_name'];
					if ($row['post_subject'] != '')
					{
						$title = ($config['sf_suppress_forum_names']) ? $row['post_subject'] : $forum_name . $user->lang['SMARTFEED_DELIMITER'] . $row['post_subject'];
					}
					else
					{
						$title = ($config['sf_suppress_forum_names']) ? 'Re: ' . $row['topic_title'] : $forum_name . $user->lang['SMARTFEED_DELIMITER'] . 'Re: ' . $row['topic_title'];
					}
					$item->title = html_entity_decode($title);		

					if ($row['topic_first_post_id'] != $row['post_id'])
					{
						if ($config['sf_show_username_in_replies'])
						{
							$title .= ($row['username'] == '') ? $user->lang['SMARTFEED_DELIMITER'] . $user->lang['SMARTFEED_REPLY_BY'] . ' ' . $user->lang['GUEST'] . ' ' . $username : $user->lang['SMARTFEED_DELIMITER'] . $user->lang['SMARTFEED_REPLY_BY'] . ' ' . $username;
						}
						else
						{
							$title .= $user->lang['SMARTFEED_DELIMITER'] . $user->lang['SMARTFEED_REPLY'];
						}
					}
					else
					{
						if ($config['sf_show_username_in_first_topic_post'])
						{
							$title .= $user->lang['SMARTFEED_DELIMITER'] . $user->lang['AUTHOR'] . ' ' . $username;
						}
					}
				}
				
				$item->title = html_entity_decode(censor_text($title));
				
				// Create a number of item attributes, not all of which are necessarily in the feed (it depends on the feed type)
				$item->date = ($config['sf_new_post_notifications_only']) ? $row['topic_last_post_time'] : $row['post_time'];
				$item->pubDate = $item->date;
				$item->link = $board_url . 'viewtopic.' . $phpEx . '?f=' . $row['forum_id'] . '&t=' . $row['topic_id'] . '&p=' . $row['post_id']  . '#p' . $row['post_id'];
				$item->source = $board_url;
				$item->category = html_entity_decode($row['forum_name']);
				$item->guid = $item->link;
				$item->descriptionHtmlSyndicated = true;
				$item->comments = $board_url . 'posting.' . $phpEx . '?mode=reply&f=' . $row['forum_id'] . '&t=' . $row['topic_id'];
				$item->author = $username;
	
				if ($feed_type==$config['sf_rss_20_value'] && !$config['sf_new_post_notifications_only']) 
				{
					// RSS 2.0 requires an email field to validate. Use a fake email address (no_email@example.com, set in feedcreator.class.php 
					// if the author field is blank) unless in user profile it says it is okay to show email address.
					$item->authorEmail = (($row['user_allow_viewemail'] == '1') && (!$config['sf_privacy_mode'])) ? $row['user_email'] : '';
				}               
				else
				{
					$item->authorEmail = '';
				}
				
				// To "dress up" the post text with bbCode, images, smilies etc., we need to use generate_text_for_display() function. This is a great
				// function. Replicating this for phpBB 2 was a pain in the buttocks!
				
				if ($config['sf_new_post_notifications_only'])
				{
					$message = $user->lang['SMARTFEED_NEW_POST_NOTIFICATIONS_ONLY'];
				}
				else
				{
					$message = censor_text($row['post_text']);
					
					$user_sig = ( $row['enable_sig'] && $row['user_sig'] != '' && $config['allow_sig'] && (!($config['sf_privacy_mode']) || $registered_user) ) ? $row['user_sig'] : '';
					
					if (($feed_style == SMARTFEED_HTML) || ($feed_style == SMARTFEED_HTMLSAFE))
					{
						$flags = (($row['enable_bbcode']) ? OPTION_FLAG_BBCODE : 0) +
							(($row['enable_smilies']) ? OPTION_FLAG_SMILIES : 0) + 
							(($row['enable_magic_url']) ? OPTION_FLAG_LINKS : 0);
							
						// If there is an image, show it. If there is a file, link to the attachment
						if ($row['post_attachment'] > 0)
						{
							$message .= sprintf("<div class=\"box\">\n<p>%s</p>\n", $user->lang['ATTACHMENTS']);
							
							// Get all attachments
							$sql3 = 'SELECT *
								FROM ' . ATTACHMENTS_TABLE . '
								WHERE post_msg_id = ' . $row['post_id'] . ' AND in_message = 0 
								ORDER BY attach_id';
							$result3 = $db->sql_query($sql3);
							while ($row3 = $db->sql_fetchrow($result3))
							{
								$file_size = round(($row3['filesize']/1024),2);
								// Show images, link to other attachments
								if (substr($row3['mimetype'],0,6) == 'image/')
								{
									$anchor_begin = '';
									$anchor_end = '';
									$post_image_text = '';
									$thumbnail_parameter = '';
									$is_thumbnail = ($row3['thumbnail'] == 1) ? true : false;
									// Logic to resize the image, if needed
									if ($is_thumbnail)
									{
										$anchor_begin = sprintf("<a href=\"%s\">", $board_url . "download/file.$phpEx?id=" . $row3['attach_id']);
										$anchor_end = '</a>';
										$post_image_text = $user->lang['SMARTFEED_POST_IMAGE_TEXT'];
										$thumbnail_parameter = '&t=1';
									}
									$message .= sprintf("%s<br /><em>%s</em> (%s KiB)<br />%s<img src=\"%s\" alt=\"%s\" title=\"%s\" />%s\n<br />%s", $row3['attach_comment'], $row3['real_filename'], $file_size, $anchor_begin, $board_url . "download/file.$phpEx?id=" . $row3['attach_id'] . $thumbnail_parameter, $row3['attach_comment'], $row3['attach_comment'], $anchor_end, $post_image_text);
								}
								else
								{
									$message .= ($row3['attach_comment'] == '') ? '' : '<em>' . $row3['attach_comment'] . '</em><br />';
									$message .= 
										sprintf("<img src=\"%s\" title=\"\" alt=\"\" /> ", 
											$board_url . 'styles/' . get_default_style() . '/imageset/icon_topic_attach.gif') .
										sprintf("<b><a href=\"%s\">%s</a></b> (%s KiB)<br />",
											$board_url . "download/file.$phpEx?id=" . $row3['attach_id'], 
											$row3['real_filename'], 
											$file_size);
								}
							}
							$db->sql_freeresult($result3);
							
							$message .= '</div>';
										
						}

						$message = generate_text_for_display($message, $row['bbcode_uid'], $row['bbcode_bitfield'], $flags);
						$message = str_replace('<img src="./', '<img src="' . $board_url, $message); 
						
						if ($user_sig != '')
						{
							$user_sig = generate_text_for_display(censor_text($user_sig), $row['user_sig_bbcode_uid'], $row['user_sig_bbcode_bitfield'], $flags);
						}
			
						$message = ($user_sig != '') ? $message . $user->lang['SMARTFEED_POST_SIGNATURE_DELIMITER'] . $user_sig : $message;
						if ($feed_style == SMARTFEED_HTMLSAFE)
						{
							$message = strip_tags($message, $allowable_tags);
						}
			
					}
					else
					{
						// Either Compact or Basic Style wanted
						if ($feed_style == SMARTFEED_BASIC)
						{
							$message = ($user_sig != '') ? $message . "\n\n" . $user_sig : $message;
						}
						$message = strip_tags($message);	// Gets rid of any embedded HTML
						strip_bbcode($message); 			// Remove the BBCode
						// Either condense all text or make line feeds explicit
						$message = ($feed_style == SMARTFEED_BASIC) ? nl2br($message) : str_replace("\n", ' ', $message);
					}
					
					// Handle the maximum number of words requested per post logic
					if ($config['sf_max_size'] > 0 && $max_word_size > 0)
					{
						$message = truncate_words($message, min($config['sf_max_size'], intval($max_word_size)));
					}
					else if ($config['sf_max_size'] > 0 && $max_word_size == 0)
					{
						$message = truncate_words($message, $config['sf_max_size']);
					}
					else if ($max_word_size > 0)
					{
						$message = truncate_words($message, intval($max_word_size));
					}
				}
				
				// Add the item (post) to the feed
				$item->description = $message;
				$rss->addItem($item);
				
			}
			
		}

		// If requested to put external items at the bottom of the feed, do it here.
		if ($config['sf_external_feeds_top'] == 0)
		{
			publish_external_feeds($feeds, $rss);
		}
		
	}
	
	// Publish the feed as RSS or Atom
	$rss->outputFeed($feed_type);
	return;

}

function base64_decode_urlsafe($input)
{

	// Thanks to phpBB forum user klapray for this logic for creating a "urlsafe" versions of base64_encode and _decode.
	return base64_decode(strtr($input, '-_.', '+/='));
	
}

function decrypt($encoded_64, $key)
{

	// This function decrypts $encoded_64 with the given $key. It uses the PHP Secure Communications Library, found
	// at http://phpseclib.sourceforge.net/. If mcrypt is compiled into PHP it will use that, otherwise it will use the functions 
	// from this library. This allows for maximum portability of Smartfeed. Thanks to Terrafrost at phpbb.com for the suggestion
	// to use this library! (He is the author.)

	global $phpbb_root_path, $phpEx;
	
	if (!defined('CRYPT_DES_MODE_3CBC'))
	{
		include($phpbb_root_path . "includes/Crypt/TripleDES.$phpEx");
	}
	$des = new Crypt_TripleDES();

	$des->setKey($key);

	// Decrypt
	$decrypted_data = $des->decrypt(base64_decode_urlsafe($encoded_64));
	return $decrypted_data;
	
}  

function get_default_style()
{

	// This function returns the default style. The default style is needed to retrive the feed
	// image.

	global $config, $db;
	
	$sql = 'SELECT template_path 
			FROM ' . STYLES_TABLE . ' s JOIN ' . STYLES_TEMPLATE_TABLE . ' t ON s.template_id = t.template_id 
			WHERE style_id = ' . $config['default_style'];
	$result = $db->sql_query($sql);
	$row = $db->sql_fetchrow($result);
	$db->sql_freeresult($result); // Query be gone!
	return $row['template_path'];
	
}  

function truncate_words($text, $max_words, $just_count_words = false)
{

	// This function returns the first $max_words from the supplied $text. If $just_count_words === true, a word count is returned. Note:
	// for consistency, HTML is stripped. This can be annoying, but otherwise HTML rendered in the feed may not be valid.
	
	global $user;
	
	if ($just_count_words)
	{
		return str_word_count(strip_tags($text));
	}
	
	$word_array = explode(' ', strip_tags($text));

	if (sizeof($word_array) <= $max_words)
	{
		return rtrim($text);
	}
	else
	{
		$truncated_text = '';
		for ($i=0; $i < $max_words; $i++) 
		{
			$truncated_text .= $word_array[$i] . ' ';
		}
		return rtrim($truncated_text) . $user->lang['SMARTFEED_MAX_WORDS_NOTIFIER'];
	}
	
}

function check_all_parents($forum_id)
{

	// This function checks all parents for a given forum_id. If any of them do not have the f_list permission
	// the function returns false, meaning the forum should not be displayed because it has a parent that should
	// not be listed. Otherwise it returns true, indicating the forum can be listed.
	
	global $db, $forum_array, $list_id;
	
	$there_are_parents = true;
	$current_forum_id = $forum_id;
	$include_this_forum = true;
	
	static $parents_loaded = false;
	static $parent_array = array();
	
	if (!$parents_loaded)
	{
		// Get a list of parent_ids for each forum and put them in an array.
		$sql = 'SELECT forum_id, parent_id 
			FROM ' . FORUMS_TABLE . '
			ORDER BY forum_id ASC';
		$result = $db->sql_query($sql);
		while ($row = $db->sql_fetchrow($result))
		{
			$parent_array[$row['forum_id']] = $row['parent_id'];
		}
		$parents_loaded = true;
		$db->sql_freeresult($result); // Query be gone!
	}
	
	while ($there_are_parents)
	{
	
		if ($parent_array[$current_forum_id] == 0) 	// No parent
		{
			$there_are_parents = false;
		}
		else
		{
			if (isset($forum_array[$parent_array[$current_forum_id]][$list_id]) && ($forum_array[$parent_array[$current_forum_id]][$list_id] == 1))
			{
				// So far so good
				$current_forum_id = $parent_array[$current_forum_id];
			}
			else
			{
				// Danger Will Robinson! No list permission exists for a parent of the requested forum, so this forum should not be shown
				$there_are_parents = false;
				$include_this_forum = false;
			}
		}
		
	}
	
	return $include_this_forum;
	
}

function publish_external_feeds($feeds, $rss)
{
	
	global $item, $config, $date_limit, $min_word_size, $max_word_size, $user;

	// If there are external feeds, publish one at a time
	foreach($feeds as $feed_url)
	{
		
		if ($feed_url <> '')
		{
			$feed = new SimplePie();
			$feed->set_feed_url($feed_url);
			$feed->enable_cache(false);
	
			// Fetch external feeds using SimplePie.
			$success = $feed->init();
			if ($success)
			{
	
				foreach ($feed->get_items(0, 0) as $feed_item)
				{
					
					if (
							(
								($date_limit == 0) || // No date range is set OR		
								($feed_item->get_date('U') >= $date_limit)	// Article is within post date range limit set
							) &&	
							(
								($min_word_size == 0) || // There is no minimum number of words to worry about OR
								(truncate_words($feed_item->get_content(), intval($max_word_size), true) >= $min_word_size) // Article has minimum words
							)	
						)
					{
						$item = new FeedItem();
						if (trim($config['sf_external_feeds_prefix']) == '')
						{
							$item->title = $feed_item->get_title();
						}
						else
						{
							$item->title = trim($config['sf_external_feeds_prefix']) . $user->lang['SMARTFEED_DELIMITER'] . $feed_item->get_title();
						}
						if (intval($max_word_size) > 0)
						{
							$item->description = truncate_words($feed_item->get_content(), intval($max_word_size));
						}
						else
						{
							$item->description = $feed_item->get_content();
						}
						
						// Create a number of item attributes, not all of which are necessarily in the feed (it depends on the feed type)
						$item->date = $feed_item->get_date();
						$item->pubDate = $feed_item->get_date();
						$item->link = $feed_item->get_permalink();
						$item->source = $feed_item->get_source();
						$item->category = $feed_item->get_category();
						$item->guid = $feed_item->get_id();
						$item->descriptionHtmlSyndicated = true;
						$item->author = $feed_item->get_author();
		
						$rss->addItem($item);
					}
				}
	
			}
			else
			{
				// Send a note to administrator there is a bad feed URL
				add_log('admin', 'LOG_CONFIG_SMARTFEED_EXTFEED', $feed_url);
			}
		}
			
	}
	
	return true;

}	

?>
